
prism.registerWidget("funnel", {

    name: "funnel",
    family: "Column",
    title: "Funnel",
    iconSmall: "/plugins/funnelWidget/widget-24.png",
    styleEditorTemplate: "/plugins/funnelWidget/styler.html",
    style: {
        isCurved: false,
        bottomPinch: 0,
        fillType: "solid",		//can also be 'gradient'
        isInverted: false,
        hoverEffects: true,
        dynamicArea: true,
        suppressZeros: true
    },
    data: {
        selection: [],
        defaultQueryResult: {},
        panels: [
            {
                name: 'Category',
                type: "series",
                itemAttributes: ["color"],
                allowedColoringTypes: function () {
                    return {
                        color: true,
                        condition: false,
                        range: false
                    }
                },
                metadata: {
                    types: ['dimensions'],
                    maxitems: 1
                },
                visibility: function (w) {
                    return w.metadata.panel("value").items.length < 2
                }
            },
            {
                name: 'value',
                type: 'visible',
                itemAttributes: ["color"],
                allowedColoringTypes: function () {
                    return {
                        color: true,
                        condition: false,
                        range: false
                    };
                },
                metadata: {
                    types: ['measures'],
                    maxitems: function (w) {
                        if (w.metadata.panel("Category").items.length == 0) {
                            return -1;
                        } else {
                            return 1;
                        }
                    }
                },
                itemAdded: function (widget, item) {
                    var colorFormatType = $$get(item, "format.color.type");
                    if ("color" === colorFormatType || "color" === colorFormatType) {
                        var color = item.format.color.color;
                        defined(color) && "transparent" != color && "white" != color && "#fff" != color && "#ffffff" != color || $jaql.resetColor(item)
                    }
                    "range" === colorFormatType && $jaql.resetColor(item), defined(item, "format.color_bkp") && $jaql.resetColor(item), defined(item, "format.members") && delete item.format.members
                }
            },
            {
                name: 'filters',
                type: 'filters',
                metadata: {
                    types: ['dimensions'],
                    maxitems: -1
                }
            }
        ],

        canColor: function (widget, panel, item) {
            return (panel.name === "value" && widget.metadata.panel("Category").items.length == 0);
            //return panel.name === "value" ;
        },

        allocatePanel: function (widget, metadataItem) {
            // measure
            if (prism.$jaql.isMeasure(metadataItem) && widget.metadata.panel("value").items.length === 0) {

                return "value";
            }
            // dimension
            else if (!prism.$jaql.isMeasure(metadataItem) && widget.metadata.panel("Category").items.length < 3) {

                return "Category";
            }
        },

        // returns true/ reason why the given item configuration is not/supported by the widget
        isSupported: function (items) {

            return this.rankMetadata(items, null, null) > -1;
        },

        // ranks the compatibility of the given metadata items with the widget
        rankMetadata: function (items, type, subtype) {

            var a = prism.$jaql.analyze(items);

            // require 1 measure, 2 dimension
            if (a.measures.length == 2 && a.dimensions.length == 1) {

                return 0;
            }

            return -1;
        },

        // populates the metadata items to the widget
        populateMetadata: function (widget, items) {

            var a = prism.$jaql.analyze(items);

            // allocating dimensions
            widget.metadata.panel("Category").push(a.dimensions);
            widget.metadata.panel("value").push(a.measures);

            // allocating filters
            widget.metadata.panel("filters").push(a.filters);
        },

        // builds a jaql query from the given widget
        buildQuery: function (widget) {

            // building jaql query object from widget metadata 
            var query = {
                datasource: widget.datasource,
                format: "json",
                isMaskedResult: true,
                metadata: []
            };

            // using categories or multiple data series?
            if (widget.metadata.panel("Category").items.length == 0) {
                // pushing all values
                for (i = 0; i < widget.metadata.panel("value").items.length; i++) {
                    query.metadata.push(widget.metadata.panel("value").items[i]);
                }

            } else {
                // Get the dimension	
                for (i = 0; i < widget.metadata.panel("Category").items.length; i++) {
                    query.metadata.push(widget.metadata.panel("Category").items[i]);
                }

                // pushing values
                for (i = 0; i < widget.metadata.panel("value").items.length; i++) {
                    query.metadata.push(widget.metadata.panel("value").items[i]);
                }
            }
            // force a sort by 
            if ($$get(query, 'metadata.1.jaql')) {
                query.metadata[1].jaql.sort = "desc";
            }

            // pushing filters
            widget.metadata.panel('filters').items.forEach(function (item) {
                item = $$.object.clone(item, true);
                item.panel = "scope";
                query.metadata.push(item);
            });
            return query;
        },

        //create highcharts data structure
        processResult: function (widget, queryResult) {
            //format data
            var newResults = [];
            var hasCategories = (widget.metadata.panels[0].items.length > 0);

            //	Function to change color of selected dim
            var lighter = function (color, luminosity) {

                // validate hex string
                color = new String(color).replace(/[^0-9a-f]/gi, '');
                if (color.length < 6) {
                    color = color[0] + color[0] + color[1] + color[1] + color[2] + color[2];
                }
                luminosity = luminosity || 0;

                // convert to decimal and change luminosity
                var newColor = "#", c, i, black = 0, white = 255;
                for (i = 0; i < 3; i++) {
                    c = parseInt(color.substr(i * 2, 2), 16);
                    c = Math.round(Math.min(Math.max(black, c + (luminosity * white)), white)).toString(16);
                    newColor += ("00" + c).substr(c.length);
                }
                return newColor;
            };

            if (hasCategories) {
                //using categories
                var categoryColors = {};

                // transform %46 into "." as it is producing errors when reading colors below
                _.forEach(_.keys(widget.metadata.panels[0].items[0].format.members), function (key) {
                    var updatedKey = key.replace(/%46/g, '.');
                    categoryColors[updatedKey] = widget.metadata.panels[0].items[0].format.members[key];
                });

                var isAnyLevelSelected = _.some(queryResult.$$rows, function (level) { return level[0].selected });
                //	Create the data objects
                for (i = 0; i < queryResult.$$rows.length; i++) {

                    var queryResultValue = $$get(queryResult, '$$rows.' + i + '.1.data');

                    //	Get the data values needed
                    var thisText = queryResult.$$rows[i][0].text,
                        thisData = _.isNumber(queryResultValue) ? queryResultValue.toString() : "1",
                        thisDataFormatted = ($$get(queryResult, '$$rows.' + i + '.1.text') || "1"),
                        thisColor = categoryColors[queryResult.$$rows[i][0].text].color;

                    //	Change the color for selected members (on highlight mode)
                    if (widget.options.dashboardFiltersMode === "select" && !queryResult.$$rows[i][0].selected && isAnyLevelSelected) {
                        thisColor = lighter(thisColor, 0.5);
                    }

                    //	Create the data array
                    newResults[i] = [thisText, thisData, thisColor, thisDataFormatted];
                }
            } else {
                // using multiple data series, instead of categories				
                for (i = 0; i < queryResult.$$rows[0].length; i++) {
                    var dataPoint = queryResult.$$rows[0][i];
                    var dataLabel = widget.metadata.panels[1].items[i].jaql.title;
                    var dataColor = widget.metadata.panels[1].items[i].format.color.color;
                    newResults[i] = [dataLabel, dataPoint.data.toString(), dataColor, dataPoint.text];
                }
            }


            return newResults;
        }
    },
    //	Check to make sure the dashboard filter doesn't filter out members when highlight
    beforequery: function (widget, event) {

        //	Is the widget set to highlight mode?
        if (widget.options.dashboardFiltersMode === "select") {

            //	Check for a dimension
            var firstItem = $$get(event, 'query.metadata.0');
            if (!firstItem) {
                return;
            }
            if (firstItem.jaql.datatype !== "numeric") {

                //	Check to see if a dashboard filter matches this dimension
                for (var i = 0; i < event.query.metadata.length; i++) {

                    //	Get this item
                    var item = event.query.metadata[i];

                    //	Found a match						
                    var needFix = (item.panel == "scope") && (item.jaql.dim == firstItem.jaql.dim)
                    if (needFix) {

                        //	Add the filter to the dimension's metadata
                        firstItem.jaql['in'] = {
                            "selected": {
                                "jaql": item.jaql
                            }
                        };

                        //	Set the inResultset attribute for the format
                        $.each(firstItem.format.members, function () {
                            this['inResultset'] = true;
                        })

                        //	Remove this item from the metadata array
                        event.query.metadata.splice(i, 1);
                    }

                }
            }
        }

    },
    render: function (s, e) {

        // Get widget elements
        var $lmnt = $(e.element);
        $lmnt.empty();

        var width = $lmnt.width(),
            height = $lmnt.height();

        // Prepare the container div
        var MyDiv = $lmnt[0],
            ObjectID = s.oid,
            ChartDivName = "Funnel-" + ObjectID;

        if (!defined(MyDiv)) {
            return;
        }

        MyDiv.setAttribute("id", ChartDivName);
        MyDiv.setAttribute("style", "width: 99%; height: 99%; margin: 0 auto");

        //	Define function for click hander
        var clicked = function (d) {
            //	Get the widget id
            var oid = $(this).closest('widget').attr('widgetid');

            //	Get the widget object
            var widget = $.grep(prism.activeDashboard.widgets.$$widgets, function (w) {
                return w.oid == oid;
            })[0];

            //	Make sure a matching widget was found and "Widget affects dashboard filters" is enabled
            if (widget && $$get(widget, "options.selector")) {

                //	Get the metadata
                var metadata = widget.rawQueryResult.metadata[0];

                //	Get the value
                var value = widget.rawQueryResult.values[d.index];

                //	Get the datasource
                var datasource = widget.datasource;

                //  Make sure that value contains values for filtering
                if ($$get(value, "0.data")) {
                    updateFilters(value);
                };

                function updateFilters(value) {

                    var filterOptions = {
                        save: true,
                        refresh: true,
                    };

                    //Create the jaql object
                    var selectedFilterJaql = {
                        jaql: {
                            dim: metadata.jaql.dim,
                            title: metadata.jaql.title,
                            table: metadata.jaql.table,
                            column: metadata.jaql.column,
                            datatype: metadata.jaql.datatype,
                            datasource: datasource,
                            filter: {
                                members: [value[0].data]
                            }
                        }
                    };

                    var isSameMemberFilterExist = _.some(prism.activeDashboard.filters.$$items,
                        function (filter) {

                            var isCascadeFilter = filter.hasOwnProperty('levels');

                            if (isCascadeFilter) {
                                return _.some(filter.levels, function (subFilterLevel) {
                                    return subFilterLevel &&
                                        subFilterLevel.dim === selectedFilterJaql.jaql.dim &&
                                        subFilterLevel.filter.members &&
                                        subFilterLevel.filter.members.toString() === selectedFilterJaql.jaql.filter.members.toString();
                                })
                            } else {
                                var isDatasourceSame = _.isMatch(filter.jaql.datasource, selectedFilterJaql.jaql.datasource);

                                return i &&
                                    filter.jaql.dim === selectedFilterJaql.jaql.dim &&
                                    filter.jaql.filter.members &&
                                    filter.jaql.filter.members.toString() === selectedFilterJaql.jaql.filter.members.toString() &&
                                    isDatasourceSame;
                            }
                        }
                    );

                    //empty filter dashboard
                    if (prism.activeDashboard.filters.length === 0) {

                        prism.activeDashboard.filters.update([selectedFilterJaql], filterOptions);
                    } else if (isSameMemberFilterExist) {

                        selectedFilterJaql.jaql.filter.multiSelection = getMultiSelectionFlag();
                        delete selectedFilterJaql.jaql.filter.members;
                        selectedFilterJaql.jaql.filter.all = true;
                        prism.activeDashboard.filters.update([selectedFilterJaql], filterOptions);
                    } else {

                        selectedFilterJaql.jaql.filter.multiSelection = getMultiSelectionFlag();
                        prism.activeDashboard.filters.update([selectedFilterJaql], filterOptions);
                    };

                    //get multiselection value from original filter
                    function getMultiSelectionFlag() {
                        var multiSelectionFlag;
                        for (var filterIndex = 0; filterIndex < prism.activeDashboard.filters.$$items.length; filterIndex++) {
                            var filter = prism.activeDashboard.filters.$$items[filterIndex];
                            var isCascadeFilter = filter.hasOwnProperty('levels');

                            if (isCascadeFilter) {
                                for (var subFilterIndex = 0; subFilterIndex < filter.levels.length; subFilterIndex++) {
                                    var subFilter = filter.levels[subFilterIndex];
                                    if (subFilter.dim === selectedFilterJaql.jaql.dim) {
                                        multiSelectionFlag = subFilter.filter.multiSelection;
                                    };
                                }
                            }
                            else {
                                if (filter.jaql.dim === selectedFilterJaql.jaql.dim) {
                                    multiSelectionFlag = filter.jaql.filter.multiSelection;
                                };
                            };

                            return multiSelectionFlag;
                        };
                    };
                };
            };

        };

        // Create options object
        var options = {
            isCurved: s.style.isCurved,
            bottomPinch: s.style.bottomPinch,
            fillType: "gradient",
            isInverted: s.style.isInverted,
            hoverEffects: true,
            dynamicArea: s.style.dynamicArea,
            suppressZeros: s.style.suppressZeros,
            onclick: clicked,
            widgetOid: e.widget.oid
        };

        // Modify PDF preview container height for better non-dynamicArea render
        if (/preview$/.test(options.widgetOid)) {

            var $overlay = $('.report-widget-overlay');
            var $headerHolder = $('.preview-widget-header-holder');

            if ($overlay[0] && $headerHolder[0]) {

                var heightDelta = $overlay.height() - $headerHolder.height();
                var printHeight = heightDelta * 0.965;

                options.height = printHeight;
            }
        }

        //	Make sure there's something to draw
        if (!$.isEmptyObject(s.queryResult)) {
            // Build chart
            var chart = new D3Funnel("#" + ChartDivName);

            if (options.isInverted) {
                // we need to clone queryResult as the method .reverse() used below changes an array it self. This
                // behaviour is error-prone because queryResult is persisted when switching between design and dashboard
                // modes which leads to reverting already reverted array, and as a result false behaviour of widget.
                var finalData = _.clone(s.queryResult);

                // invert the data
                chart.draw(finalData.reverse(), options);
            } else {
                // Otherwise, just use the regular data
                chart.draw(s.queryResult, options);
            }

        }


    },


    destroy: function (s, e) { }
});